mod actor;
mod event_loop;
mod request;
mod section_runner_json;
mod topics;
mod update_listener;

pub use request::RequestContext;
pub use update_listener::UpdateListener;

use self::topics::Topics;
use section_runner_json::SecRunnerStateJson;
use sprinklers_actors::section_runner::SecRunnerState;
use sprinklers_core::model::{Program, ProgramId, Programs, Section, SectionId, Sections};

use actix::{Actor, Addr};
use eyre::WrapErr;
use rumqttc::{LastWill, MqttOptions, QoS};
use std::{
    ops::{Deref, DerefMut},
    sync::Arc,
};

#[derive(Clone, Debug)]
pub struct Options {
    pub broker_host: String,
    pub broker_port: u16,
    pub device_id: String,
    pub client_id: String,
}

#[derive(Clone)]
pub struct MqttInterface {
    client: rumqttc::AsyncClient,
    topics: Topics<Arc<str>>,
}

impl MqttInterface {
    fn new(options: Options) -> (Self, rumqttc::EventLoop) {
        let mqtt_prefix = format!("devices/{}", options.device_id);
        let topics: Topics<Arc<str>> = Topics::new(mqtt_prefix.into());
        let mut mqtt_opts =
            MqttOptions::new(options.client_id, options.broker_host, options.broker_port);

        let last_will = LastWill::new(topics.connected(), "false", QoS::AtLeastOnce, true);
        mqtt_opts.set_last_will(last_will);

        let (client, event_loop) = rumqttc::AsyncClient::new(mqtt_opts, 16);

        (Self { client, topics }, event_loop)
    }

    async fn publish_data<P>(&mut self, topic: String, payload: &P) -> eyre::Result<()>
    where
        P: serde::Serialize,
    {
        let payload_vec =
            serde_json::to_vec(payload).wrap_err("failed to serialize publish payload")?;
        self.client
            .publish(topic, QoS::AtLeastOnce, true, payload_vec)
            .await
            .wrap_err("failed to publish")?;
        Ok(())
    }

    async fn publish_connected(&mut self, connected: bool) -> eyre::Result<()> {
        self.publish_data(self.topics.connected(), &connected)
            .await
            .wrap_err("failed to publish connected topic")
    }

    async fn cancel(&mut self) -> Result<(), rumqttc::ClientError> {
        self.client.cancel().await
    }

    pub async fn publish_sections(&mut self, sections: &Sections) -> eyre::Result<()> {
        let section_ids: Vec<_> = sections.keys().cloned().collect();
        self.publish_data(self.topics.sections(), &section_ids)
            .await
            .wrap_err("failed to publish section ids")?;
        for section in sections.values() {
            self.publish_section(section).await?;
        }
        Ok(())
    }

    pub async fn publish_section(&mut self, section: &Section) -> eyre::Result<()> {
        self.publish_data(self.topics.section_data(section.id), section)
            .await
            .wrap_err("failed to publish section")
    }

    // Section state can be derived from section runner state...
    pub async fn publish_section_state(
        &mut self,
        section_id: SectionId,
        state: bool,
    ) -> eyre::Result<()> {
        self.publish_data(self.topics.section_state(section_id), &state)
            .await
            .wrap_err("failed to publish section state")
    }

    pub async fn publish_programs(&mut self, programs: &Programs) -> eyre::Result<()> {
        let program_ids: Vec<_> = programs.keys().cloned().collect();
        self.publish_program_ids(&program_ids).await?;
        for program in programs.values() {
            self.publish_program(program).await?;
        }
        Ok(())
    }

    pub async fn publish_program_ids(&mut self, program_ids: &[ProgramId]) -> eyre::Result<()> {
        self.publish_data(self.topics.programs(), &program_ids)
            .await
            .wrap_err("failed to publish program ids")?;
        Ok(())
    }

    pub async fn publish_program(&mut self, program: &Program) -> eyre::Result<()> {
        self.publish_data(self.topics.program_data(program.id), &program)
            .await
            .wrap_err("failed to publish program")
    }

    pub async fn publish_programs_diff(
        &mut self,
        old_programs: &Programs,
        programs: &Programs,
    ) -> eyre::Result<()> {
        for (id, program) in programs {
            let publish = match old_programs.get(id) {
                Some(old_program) => !Arc::ptr_eq(old_program, program),
                None => {
                    let program_ids: Vec<_> = programs.keys().cloned().collect();
                    self.publish_program_ids(&program_ids).await?;
                    true
                }
            };
            if publish {
                self.publish_program(program).await?;
            }
        }
        Ok(())
    }

    pub async fn publish_program_running(
        &mut self,
        program_id: ProgramId,
        running: bool,
    ) -> eyre::Result<()> {
        self.publish_data(self.topics.program_running(program_id), &running)
            .await
            .wrap_err("failed to publish program running")
    }

    pub async fn publish_program_next_run(
        &mut self,
        program_id: ProgramId,
        next_run: chrono::DateTime<chrono::Local>,
    ) -> eyre::Result<()> {
        let payload = next_run.to_rfc3339();
        self.publish_data(self.topics.program_next_run(program_id), &payload)
            .await
            .wrap_err("failed to publish program next run")
    }

    pub async fn publish_section_runner(&mut self, sr_state: &SecRunnerState) -> eyre::Result<()> {
        let json: SecRunnerStateJson = sr_state.into();
        self.publish_data(self.topics.section_runner(), &json)
            .await
            .wrap_err("failed to publish section runner")
    }

    async fn publish_response(&mut self, resp: request::ResponseWithId) -> eyre::Result<()> {
        let payload_vec =
            serde_json::to_vec(&resp).wrap_err("failed to serialize request response")?;
        // TODO: if couldn't serialize, just in case can have a static response
        self.client
            .publish(self.topics.responses(), QoS::AtMostOnce, false, payload_vec)
            .await
            .wrap_err("failed to publish request response")?;
        Ok(())
    }

    async fn subscribe_requests(&mut self) -> eyre::Result<()> {
        self.client
            .subscribe(self.topics.requests(), QoS::ExactlyOnce)
            .await?;
        Ok(())
    }
}

pub struct MqttInterfaceTask {
    interface: MqttInterface,
    addr: Addr<actor::MqttActor>,
}

impl MqttInterfaceTask {
    pub fn start(options: Options, request_context: RequestContext) -> Self {
        let (interface, event_loop) = MqttInterface::new(options);

        let addr = actor::MqttActor::new(interface.clone(), event_loop, request_context).start();

        Self { interface, addr }
    }

    pub async fn quit(self) -> eyre::Result<()> {
        self.addr.send(actor::Quit).await?;
        Ok(())
    }
}

impl Deref for MqttInterfaceTask {
    type Target = MqttInterface;

    fn deref(&self) -> &Self::Target {
        &self.interface
    }
}

impl DerefMut for MqttInterfaceTask {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.interface
    }
}
