#![warn(clippy::all)]
#![warn(clippy::print_stdout)]

// mod option_future;
mod state_manager;

use sprinklers_actors as actors;
use sprinklers_core::section_interface::MockSectionInterface;
use sprinklers_database as database;
use sprinklers_mqtt as mqtt;

use eyre::Result;
use std::sync::Arc;
use tracing::{debug, info};
use tracing_subscriber::EnvFilter;

#[actix_rt::main]
async fn main() -> Result<()> {
    tracing_subscriber::fmt()
        .with_ansi(true)
        .with_env_filter(
            EnvFilter::try_from_default_env().unwrap_or_else(|_| EnvFilter::new("info")),
        )
        .init();
    info!("Starting sprinklers_rs...");
    color_eyre::install()?;

    let db_conn = database::setup_db()?;

    let sections = database::query_sections(&db_conn)?;
    for sec in sections.values() {
        debug!(section = debug(&sec), "read section");
    }

    // TODO: Section interface which actual does something. Preferrably selectable somehow
    let section_interface: Arc<_> = MockSectionInterface::new(6).into();
    let mut section_runner = actors::SectionRunner::new(section_interface);
    let mut program_runner = actors::ProgramRunner::new(section_runner.clone());

    let state_manager = crate::state_manager::StateManagerThread::start(db_conn);

    let mqtt_options = mqtt::Options {
        broker_host: "localhost".into(),
        broker_port: 1883,
        device_id: "sprinklers_rs-0001".into(),
        client_id: "sprinklers_rs-0001".into(),
    };
    // TODO: have ability to update sections / other data
    let request_context = mqtt::RequestContext {
        sections: sections.clone(),
        section_runner: section_runner.clone(),
        program_runner: program_runner.clone(),
        state_manager: state_manager.clone(),
    };
    let mut mqtt_interface = mqtt::MqttInterfaceTask::start(mqtt_options, request_context);

    let mut update_listener = mqtt::UpdateListener::start(mqtt_interface.clone());
    update_listener.listen_section_events(section_runner.subscribe().await?);
    update_listener.listen_section_runner(section_runner.get_state_recv());
    update_listener.listen_programs(state_manager.get_programs());
    update_listener.listen_program_events(program_runner.subscribe().await?);

    // Only listen to programs now so above subscriptions get events
    program_runner.listen_programs(state_manager.get_programs());

    program_runner.update_sections(sections.clone()).await?;
    // TODO: update listener should probably do this
    mqtt_interface.publish_sections(&sections).await?;
    for section_id in sections.keys() {
        mqtt_interface
            .publish_section_state(*section_id, false)
            .await?;
    }

    info!("sprinklers_rs initialized");

    tokio::signal::ctrl_c().await?;
    info!("Interrupt received, shutting down");

    update_listener.quit().await?;
    mqtt_interface.quit().await?;
    drop(state_manager);
    program_runner.quit().await?;
    section_runner.quit().await?;
    actix::System::current().stop();

    Ok(())
}
